from .fileformat import format_handlers  # noqa: F401

from .interface import (  # noqa: F401
    HelpAction,
    NumericRangeType,
    FileAccessType,
    DirectoryAccessType,
)

from .io import (  # noqa: F401
    NamedBytesIO,
    NamedStringIO,
)

from .iteration import (  # noqa: F401
    prime_iterator,
    storeresults,
)



"""Pipeline Step to extract networking traffic via nPrint: Net"""
import argparse
import collections
import concurrent.futures
import itertools
import os
import pathlib
import re
import sys
import textwrap
import typing

import nprintml
from nprintml import pipeline
from nprintml.util import (
    DirectoryAccessType,
    FileAccessType,
    HelpAction,
    NamedBytesIO,
    prime_iterator,
)

from . import execute


NPRINT_QUEUE_MAX = 10 ** 6


class NetResult(typing.NamedTuple):
    """Pipeline Step results for Net"""
    nprint_stream: typing.Generator[NamedBytesIO, None, None]


class Net(pipeline.Step):
    """Extend given `ArgumentParser` with nPrint interface and invoke
    `nprint` command to initiate nprintML pipeline.

    Returns a `NetResult`.

    """
    __provides__ = NetResult
    __requires__ = ('labels',)

    def __init__(self, parser):
        self.args = None  # set by __pre__

        self.group_parser = parser.add_argument_group(
            "extraction of features from network traffic via nPrint",

            "(full information can be found at https://nprint.github.io/nprint/)"
        )

        # record of subset of arguments NOT passed to nprint
        self.own_arguments = set()

        self.group_parser.add_argument(
            '-4', '--ipv4',
            action='store_true',
            help="include ipv4 headers",
        )
        self.group_parser.add_argument(
            '-6', '--ipv6',
            action='store_true',
            help="include ipv6 headers",
        )
        self.group_parser.add_argument(
            '-A', '--absolute-timestamps', '--absolute_timestamps',
            action='store_true',
            help="include absolute timestamp field",
        )
        self.group_parser.add_argument(
            '-c', '--count',
            metavar='INTEGER',
            type=int,
            help="number of packets to parse (if not all)",
        )
        self.group_parser.add_argument(
            '-C', '--csv-file', '--csv_file',
            type=FileAccessType(os.R_OK),
            metavar='FILE',
            help="csv (hex packets) infile",
        )
        self.group_parser.add_argument(
            '-d', '--device',
            help="device to capture from if live capture",
        )
        self.group_parser.add_argument(
            '-e', '--eth',
            action='store_true',
            help="include eth headers",
        )
        self.group_parser.add_argument(
            '-f', '--filter',
            help="filter for libpcap",
        )
        self.group_parser.add_argument(
            '-i', '--icmp',
            action='store_true',
            help="include icmp headers",
        )
        self.group_parser.add_argument(
            '-I', '--ip-file', '--ip_file',
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="file of IP addresses to filter with (1 per line), "
                 "can be combined with num_packets for num_packets per ip",
        )
        self.group_parser.add_argument(
            '-N', '--nprint-file', '--nPrint_file',
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="nPrint infile",
        )
        self.group_parser.add_argument(
            '-O', '--write-index', '--write_index',
            choices=range(6),
            metavar='INTEGER',
            type=int,
            help=textwrap.dedent("""\
                output file index (first column)
                select from:
                    0: source IP (default)
                    1: destination IP
                    2: source port
                    3: destination port
                    4: flow (5-tuple)
                    5: wlan tx mac"""),
        )
        self.group_parser.add_argument(
            '-p', '--payload',
            metavar='INTEGER',
            type=int,
            help="include n bytes of payload",
        )
        self.group_parser.add_argument(
            '-P', '--pcap-file', '--pcap_file',
            action='append',
            default=[],
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="pcap infile",
        )
        self.own_arguments.add('pcap_file')
        self.group_parser.add_argument(
            '--pcap-dir', '--pcap_dir',
            action='append',
            default=[],
            metavar='DIR',
            type=DirectoryAccessType(ext='.pcap'),
            help="directory containing pcap infile(s) with file extension '.pcap'",
        )
        self.own_arguments.add('pcap_dir')
        self.group_parser.add_argument(
            '-r', '--radiotap',
            action='store_true',
            help="include radiotap headers",
        )
        self.group_parser.add_argument(
            '-R', '--relative-timestamps', '--relative_timestamps',
            action='store_true',
            help="include relative timestamp field",
        )
        self.group_parser.add_argument(
            '-t', '--tcp',
            action='store_true',
            help="include tcp headers",
        )
        self.group_parser.add_argument(
            '-u', '--udp',
            action='store_true',
            help="include udp headers",
        )
        self.group_parser.add_argument(
            '-w', '--wlan',
            action='store_true',
            help="include wlan headers",
        )
        self.group_parser.add_argument(
            '-x', '--nprint-filter', '--nprint_filter',
            metavar='STRING',
            help="regex to filter bits out of nPrint output "
                 "(for details see --help-nprint-filter)",
        )
        self.group_parser.add_argument(
            '--help-nprint-filter', '--nprint-filter-help', '--nprint_filter_help',
            action=HelpAction,
            help_action=lambda *_parser_args: execute.nprint('--nprint_filter_help'),
            help="describe regex possibilities and exit",
        )

        self.group_parser.add_argument(
            '--save-nprint',
            action='store_true',
            help="save nPrint output(s) to disk (default: not saved)",
        )
        self.own_arguments.add('save_nprint')

    output_tag = 'nprint'
    default_output_name = 'netcap.npt'

    @property
    def output_directory(self):
        return self.args.outdir / self.output_tag

    def ensure_output_directory(self):
        if self.args.save_nprint:
            self.output_directory.mkdir()

    @classmethod
    def get_output_name(cls, pcap_output):
        if pcap_output is None:
            return cls.default_output_name

        # pcap_output expects str and re is fast
        return cls.get_output_name.pattern.sub('.npt', pcap_output)

    get_output_name.__func__.pattern = re.compile(r'\.pcap$', re.I)

    def make_output_path(self, npt_file):
        npt_path = self.output_directory / npt_file

        if npt_path.exists():
            raise FileExistsError(None, 'nPrint output path collision', str(npt_path))

        npt_path.parent.mkdir(parents=True, exist_ok=True)

        return npt_path

    def generate_pcaps(self):
        if self.args.pcap_file or self.args.pcap_dir:
            # stream pair of pcap path & short ref path for reconstructing tree
            for pcap_file in self.args.pcap_file:
                pcap_path = pathlib.Path(pcap_file)
                yield (pcap_path, pcap_path.name)

            for pcap_dir in self.args.pcap_dir:
                for pcap_path in pcap_dir.rglob('*.pcap'):
                    pcap_file = pcap_path.relative_to(pcap_dir)
                    yield (pcap_path, str(pcap_file))

            return

        yield (None, None)

    def generate_argv(self, pcap_file=None, npt_file=None):
        """Construct arguments for `nprint` command."""
        # generate shared/global arguments
        if self.args.verbosity >= 3:
            yield '--verbose'

        # support arbitrary pcap infile(s)
        if pcap_file:
            yield from ('--pcap_file', pcap_file)

        # add group (nPrint-specific) arguments
        for action in self.group_parser._group_actions:
            if action.dest in self.own_arguments:
                continue

            try:
                value = getattr(self.args, action.dest)
            except AttributeError:
                if action.default == argparse.SUPPRESS:
                    continue

                raise

            key = action.option_strings[-1]

            if value is not action.default:
                yield key

                if not isinstance(value, bool):
                    yield str(value)

        # add output path
        if npt_file:
            yield from ('--write_file', npt_file)

    def write_meta(self, target):
        args = ' '.join(self.generate_argv('[input_pcap]'))
        target[self.output_tag] = {'cmd': f'nprint {args}'}

    def filtermap_pcaps(self, pcap_files, labels=None):
        skipped_files = collections.deque(maxlen=4)
        skipped_count = 0

        for (pcap_path, pcap_file) in pcap_files:
            if pcap_path:
                if labels is not None and pcap_file not in labels.index:
                    skipped_files.append(pcap_file)
                    skipped_count += 1
                    continue

                npt_file = self.get_output_name(pcap_file)

                yield (pcap_path, npt_file)
            else:
                npt_file = self.get_output_name(None)

                yield (None, npt_file)

        if skipped_count > 0 and self.args.verbosity >= 1:
            print('Skipped', skipped_count, 'PCAP file(s) missing from labels file:')

            for skipped_file in skipped_files:
                print(f'\t{skipped_file}')

            if skipped_count > len(skipped_files):
                print('\t...')

    def execute_nprint(self, pcap_file, npt_file):
        result = execute.nprint(
            *self.generate_argv(pcap_file),
            stdout=execute.nprint.PIPE,
        )

        if self.args.save_nprint:
            out_path = self.make_output_path(npt_file)
            out_path.write_bytes(result.stdout)

        return NamedBytesIO(result.stdout, name=npt_file)

    def generate_npts(self, file_stream, timing):
        # Spawn thread pool of same size as number of concurrent
        # subprocesses we would like to maintain.
        with timing, concurrent.futures.ThreadPoolExecutor(
            max_workers=self.args.concurrency,
            thread_name_prefix=__name__,
        ) as executor:
            # Enqueue/schedule at most NPRINT_QUEUE_MAX nprint calls
            # (not all up front, in case this is VERY large, to spare RAM).
            futures = {
                executor.submit(self.execute_nprint, *args)
                for args in itertools.islice(file_stream, NPRINT_QUEUE_MAX)
            }

            if self.args.verbosity >= 3:
                print('Enqueued', len(futures), 'nPrint task(s)',
                      'to be processed by at most', self.args.concurrency, 'worker(s)')

            all_done = False
            failure_count = 0

            while futures:
                # Wait only long enough for at least one call to complete.
                (completed, futures) = concurrent.futures.wait(
                    futures,
                    return_when=concurrent.futures.FIRST_COMPLETED,
                )

                added_count = 0

                for future in completed:
                    # Top off the queue with at most one remaining call
                    # for each completed call.
                    for (added_count, args) in enumerate(itertools.islice(file_stream, 1),
                                                         added_count + 1):
                        futures.add(executor.submit(self.execute_nprint, *args))

                    if not all_done and added_count == 0 and not futures:
                        all_done = True

                        if self.args.verbosity >= 3:
                            print('nPrint tasks complete:', timing.time_elapsed)

                    # Share the result.
                    try:
                        yield future.result()
                    except execute.nprint.CommandError:
                        failure_count += 1

                if added_count and self.args.verbosity >= 3:
                    print('Enqueued', added_count, 'more nPrint task(s)')

            if failure_count > 0:
                print('nPrint task(s) failed:', failure_count)

    def __pre__(self, parser, args, results):
        try:
            warn_version_mismatch()
        except execute.nprint.NoCommand:
            parser.error("nprint command could not be found on PATH "
                         "(to install see nprint-install)")

        self.args = args

    def __call__(self, args, results):
        self.write_meta(results.meta)

        self.ensure_output_directory()

        pcap_files = self.generate_pcaps()

        file_stream = self.filtermap_pcaps(pcap_files, results.labels)

        # Time stream specially as it continues even after step formally completed:
        stream_timing = results.__timing_steps__[self.generate_npts] = pipeline.Timing()

        npt_stream = self.generate_npts(file_stream, stream_timing)

        # Initialize results generator eagerly; (it will handle own buffer):
        active_stream = prime_iterator(npt_stream)

        return NetResult(
            nprint_stream=active_stream,
        )


def warn_version_mismatch():
    """Warn if nPrint intended version doesn't match what's on PATH."""
    version_result = execute.nprint('--version', stdout=execute.nprint.PIPE)
    version_output = version_result.stdout.decode()
    version_match = re.match(r'nprint ([.\d]+)', version_output)
    if version_match:
        version_installed = version_match.group(1)
        if version_installed != nprintml.__nprint_version__:
            command_path = version_result.args[0]
            print(
                "[warn]",
                f"nprint expected version for nprintML ({nprintml.__nprint_version__}) "
                f"does not match version on PATH ({version_installed} at {command_path})",
                file=sys.stderr,
            )
    else:
        print(
            "[warn]",
            f"failed to parse version of nprint installed ({version_output})",
            file=sys.stderr,
        )



import io


class _NamedIO:

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.name}>'


class NamedStringIO(_NamedIO, io.StringIO):
    """StringIO featuring a `name` attribute to reflect the path
    represented by its contents.

    """
    def __init__(self, initial_value='', newline='\n', name=None):
        super().__init__(initial_value, newline)
        self.name = name


class NamedBytesIO(_NamedIO, io.BytesIO):
    """BytesIO featuring a `name` attribute to reflect the path
    represented by its contents.

    """
    def __init__(self, initial_bytes=b'', name=None):
        super().__init__(initial_bytes)
        self.name = name



import unittest

import pandas as pd

from nprintml.util import format_handlers

from .base import mktestdir


class TestFileFormat(unittest.TestCase):

    def test_io(self):
        for write_format in (
            'csv.gz',
            'feather.zstd',
            'parquet.snappy',
        ):
            with self.subTest(write_format=write_format), \
                 mktestdir() as tempdir:

                features = pd.DataFrame(
                    {
                        'colA': [0] * 4,
                        'colB': [1] * 4,
                        'colC': [0] * 4,
                    },
                    index=['a', 'a', 'b', 'c'],
                )

                writer = format_handlers.get_writer(write_format)
                outpath = writer(features, tempdir)

                reader = format_handlers.get_reader(outpath)
                result = reader(outpath)

                pd.testing.assert_frame_equal(
                    result,
                    features,
                    check_exact=True,   # values should be identical
                    check_names=False,  # don't worry about the index "name"
                )



"""Tests for the nprintML Pipeline"""
import argparse
import unittest

from nprintml import cli, pipeline

import nprintml.net.step
import nprintml.label.step
import nprintml.learn.step


class TestPipeline(unittest.TestCase):

    def setUp(self):
        self.parser = cli.build_parser()

    def test_eager_ordering(self):
        "pipeline determines step ordering upon inst'n and so extends parser"
        self.assertEqual(len(self.parser._action_groups), 3)

        pline = pipeline.Pipeline(self.parser)

        self.assertEqual(len(self.parser._action_groups), 6)

        for (defining_word, action_group) in zip(
            ('network', 'label', 'automl'),
            self.parser._action_groups[3:],
        ):
            self.assertIn(defining_word, action_group.title)

        pipeline_steps = [type(step) for step in pline]
        self.assertSequenceEqual(pipeline_steps, (
            nprintml.net.step.Net,
            nprintml.label.step.Label,
            nprintml.learn.step.Learn,
        ))



import abc
import pathlib

import pandas as pd


class FormatRegistry(dict):

    def register(self, handler, name, *aliases):
        for alias in (name,) + aliases:
            self[alias] = handler

    def get_writer(self, full_format):
        dot_count = full_format.count('.')

        if dot_count <= 1:
            (file_format, file_compression) = (full_format.split('.') if dot_count == 1
                                               else (full_format, None))

            try:
                handler = self[file_format]
            except KeyError:
                pass
            else:
                return handler(file_compression).write

        raise NotImplementedError(full_format)

    def get_reader(self, path):
        file_path = path if isinstance(path, pathlib.Path) else pathlib.Path(path)

        suffixes = (suffix.lower().lstrip('.') for suffix in reversed(file_path.suffixes))

        compression = None

        for suffix in suffixes:
            try:
                handler = self[suffix]
            except KeyError:
                if compression is None:
                    compression = suffix
                else:
                    break
            else:
                return handler(compression).read

        raise NotImplementedError(path)


format_handlers = FormatRegistry()


class FormatHandler(abc.ABC):

    __fileformats__ = format_handlers

    @classmethod
    def __init_subclass__(cls, *, names=(), **kwargs):
        """Register handler upon declaration."""
        super().__init_subclass__(**kwargs)

        if names:
            cls.__fileformats__.register(cls, *names)
        else:
            cls.__fileformats__.register(cls, cls.__name__.lower())

    def __init__(self, compression=None):
        self.compression = compression

    @abc.abstractmethod
    def read(self, path):
        pass

    @abc.abstractmethod
    def write(self, data, outdir):
        pass


class Csv(FormatHandler):

    def read(self, path):
        return pd.read_csv(path, index_col=0)

    def write(self, data, outdir):
        outname = 'features.csv'

        if self.compression:
            outname = f'{outname}.{self.compression}'

        outdir = outdir if isinstance(outdir, pathlib.Path) else pathlib.Path(outdir)
        outpath = outdir / outname

        # Unlike others to_csv infers compression from name
        data.to_csv(outpath)

        return outpath


class Parquet(FormatHandler, names=('parquet', 'parq')):

    def read(self, path):
        return pd.read_parquet(path)

    def write(self, data, outdir):
        outdir = outdir if isinstance(outdir, pathlib.Path) else pathlib.Path(outdir)
        outpath = outdir / 'features.parq'

        data.to_parquet(outpath, compression=self.compression)

        return outpath


class Feather(FormatHandler, names=('feather', 'fhr')):

    def read(self, path):
        data = pd.read_feather(path)

        if len(data.columns) > 0:
            column_name = data.columns[0]
            data.set_index(column_name, inplace=True)

        return data

    def write(self, data, outdir):
        outdir = outdir if isinstance(outdir, pathlib.Path) else pathlib.Path(outdir)
        outpath = outdir / 'features.fhr'

        # Like CSV, (unlike Parquet), Feather does not support custom indices.
        #
        # (And, unlike to_csv, to_feather raises an error if you mess this up.)
        #
        data.reset_index().to_feather(outpath, compression=self.compression)

        return outpath



import pathlib

import numpy as np
import pandas as pd

from . import LabelAggregator, AggregationLengthError, AggregationPathError


class IndexLabelAggregator(LabelAggregator):
    """A LabelAggregator to aggregates labels according to the *index*
    of nPrint output.

    This can be the src / dst ip, the src / dst port, or an entire flow
    label.

    """
    @classmethod
    def normalize_npt(cls, npt_csv):
        if isinstance(npt_csv, str):
            return cls.normalize_npt(pathlib.Path(npt_csv))

        if isinstance(npt_csv, pathlib.Path):
            if npt_csv.is_dir():
                try:
                    (npt_csv,) = npt_csv.iterdir()
                except ValueError:
                    raise AggregationPathError(
                        f"{cls.__name__} expects exactly one nPrint data file but "
                        f"specified directory contains none or more than one: '{npt_csv}'"
                    )

            return npt_csv

        # treat as stream
        try:
            (npt_csv,) = npt_csv
        except ValueError:
            raise AggregationLengthError(
                f"{cls.__name__} expects exactly one nPrint data result but "
                f"specified stream contains none or more than one: {npt_csv}"
            )
        else:
            return npt_csv

    def __call__(self, npt_csv, compress=False, sample_size=1, path_input_base=None):
        """Generate index driven features by loading an nPrint result
        and attaching labels, grouping by a given sample size if
        desired.

        `npt_csv` may specify either a path to an nPrint result (`str`
        or `pathlib.Path`) OR a stream containing an open file-like
        object (exposing both `read()` and `name`).

        (If a path, `npt_csv` may specify a directory, so long as the
        directory contains only one nPrint result file.)

        """
        npt_input = self.normalize_npt(npt_csv)

        print('Loading nPrint:', self.filerepr(npt_input))
        npt = self.load_npt(npt_input)

        print('Loaded 1 nprint')
        print('  nPrint shape:', npt.shape)

        if compress:
            print('Compressing nPrint')
            npt = self.compress_npt(npt)
            print('  compressed nPrint shape:', npt.shape)

        print('Loading labels:', self.filerepr(self.label_csv))
        self.labels = self.load_label(self.label_csv)
        print('  number of labels:', self.labels.shape[0])

        if sample_size > 1:
            print('Grouping by sample size')
            npt = self.regroup_npt(npt, sample_size)
            print('  New shape of dataframe:', npt.shape)

        print('Attaching labels to nPrints')
        (npt, missing_labels, ogns, nns) = self.attach_label(npt, self.labels)
        print('  labels attached: missing labels for:')
        for missing_label in missing_labels:
            print('    {0}'.format(missing_label))
        print('    missing labels caused samples to be dropped:', (ogns - nns))

        return npt

    def regroup_npt(self, npt, sample_size):
        """Take a large nPrint dataframe, groups it by the index, then
        split each separated sample into n sample_sized samples.

        """
        # If the sample size is 1, its just per packet, which we already have
        if sample_size <= 1:
            return npt

        grouped = []
        indexes = []

        for (lid, group) in npt.groupby(npt.index):
            # Pad rows to get an even division for sample size
            pad_rows = sample_size - group.shape[0] % sample_size
            fill_row = [-1] * group.shape[1]
            for index in range(group.shape[0], group.shape[0] + pad_rows):
                group.loc[index] = fill_row

            # actually separate group into samples and flatten for 1d learning
            num_splits = group.shape[0] / sample_size
            for sample in np.split(group, num_splits):
                flattened_sample = sample.to_numpy().flatten()
                grouped.append(flattened_sample)
                indexes.append(lid)

        cols_flat = self.flatten_columns(npt.columns, sample_size)
        return pd.DataFrame(grouped, index=indexes, columns=cols_flat)



import argparse
import os
import pathlib


class HelpAction(argparse.Action):
    """Pluggable "help" action.

    Based on the built-in "help" action, enabling the definition of
    additional "--help-*" flags.

    Help output must be printed by the given callable `help_action`.

    When the flag is set, `help_action` is invoked and then the process
    is exited.

    """
    def __init__(self,
                 option_strings,
                 *,
                 help_action,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

        self.help_action = help_action

    def __call__(self, parser, namespace, values, option_string=None):
        self.help_action(parser, namespace, values, option_string)
        parser.exit()


class NumericRangeType:
    """Argument type validating that given value is of configured
    numeric type and bounds.

    Bounds may be specified as either `list` or `tuple`, the convention
    indicating whether bounds are "inclusive" or "non-inclusive",
    respectively.

    Either the lower or upper bound may be specified as `None`,
    indicating no bound.

    For example:

        parser.add_argument(
            '--test-size',
            default=0.3,
            metavar='FLOAT',
            type=NumericRangeType(float, (0, 1)),
        )

    The above argument, `--test-size`, will cast its input to `float`,
    (or print an error for non-float input). Moreover, a "not in range"
    error will be printed for inputs equal to zero or equal to or
    greater than one.

        parser.add_argument(
            '--threads',
            default=1,
            metavar='INTEGER',
            type=NumericRangeType(int, (0, None)),
        )

    In the above example, inputs are instead enforced as `int`; and,
    there is _only_ a lower bound -- inputs must only be greater than
    zero.

        parser.add_argument(
            '-q', '--quality',
            default=0,
            metavar='INTEGER',
            type=NumericRangeType(int, [0, 4]),
        )

    Arguments may also be given inclusive bounds, as above -- inputs to
    this example must be greater than or equal to zero, and less than or
    equal to four.

    """
    bounds_message = "upper and lower bounds must be list or tuple of two elements"

    def __init__(self, numeric_type, bounds):
        self.numeric_type = numeric_type

        try:
            (self.lower_bound, self.upper_bound) = bounds
        except ValueError as exc:
            raise ValueError(f"{self.bounds_message} not: {bounds!r}") from exc
        except TypeError as exc:
            raise TypeError(
                f"{self.bounds_message} not {bounds.__class__.__name__}: {bounds!r}"
            ) from exc

        if isinstance(bounds, tuple):
            self.inclusive = False
        elif isinstance(bounds, list):
            self.inclusive = True
        else:
            raise TypeError(f"{self.bounds_message} not {bounds.__class__.__name__}: {bounds!r}")

    def __call__(self, value):
        try:
            number = self.numeric_type(value)
        except ValueError:
            raise argparse.ArgumentTypeError(f"not {self.numeric_type.__name__}: {value!r}")

        if self.inclusive:
            if (
                (self.lower_bound is not None and number < self.lower_bound) or
                (self.upper_bound is not None and number > self.upper_bound)
            ):
                raise argparse.ArgumentTypeError(
                    f"not in range [{self.lower_bound}, {self.upper_bound}]: {number!r}"
                )
        else:
            if (
                (self.lower_bound is not None and number <= self.lower_bound) or
                (self.upper_bound is not None and number >= self.upper_bound)
            ):
                raise argparse.ArgumentTypeError(
                    f"not in range ({self.lower_bound}, {self.upper_bound}): {number!r}"
                )

        return number


class FileAccessType:
    """Argument type to test a supplied filesystem path for specified
    access.

    Access level is indicated by bit mask.

    `argparse.FileType` may be preferred when the path should be opened
    in-process. `FileAccessType` allows for greater flexibility -- such
    as passing the path on to a subprocess -- while still validating
    access to the path upfront.

    """
    modes = {
        os.X_OK: 'execute',
        os.W_OK: 'write',
        os.R_OK: 'read',
        os.R_OK | os.X_OK: 'read-execute',
        os.R_OK | os.W_OK: 'read-write',
        os.R_OK | os.W_OK | os.X_OK: 'read-write-execute',
    }

    def __init__(self, access):
        self.access = access

        if access not in self.modes:
            raise ValueError("bad mask", access)

    @property
    def mode(self):
        return self.modes[self.access]

    def __call__(self, path):
        if os.path.isfile(path) and os.access(path, self.access):
            return path

        raise argparse.ArgumentTypeError(f"can't access file at '{path}' ({self.mode})")


class DirectoryAccessType:
    """Argument type to test a supplied filesystem directory path."""

    def __init__(self, *, ext='', exists=None, empty=False, non_empty=False):
        if ext:
            non_empty = True

        if non_empty:
            exists = True

        if empty and non_empty:
            raise TypeError("directory cannot be both empty and non-empty")

        self.ext = ext
        self.exists = exists
        self.empty = empty
        self.non_empty = non_empty

    def __call__(self, value):
        path = pathlib.Path(value)

        if self.exists is not None:
            if self.exists:
                if not path.is_dir():
                    raise argparse.ArgumentTypeError(f"no such directory '{value}'")
            else:
                if path.exists():
                    raise argparse.ArgumentTypeError(f"path already exists '{value}'")

                if not os.access(path.parent, os.W_OK):
                    raise argparse.ArgumentTypeError(f"path not write-accessible '{value}'")

        if self.empty and any(path.glob('*')):
            raise argparse.ArgumentTypeError(f"directory is not empty '{value}'")

        if self.non_empty:
            count = 0
            for (count, child) in enumerate(path.rglob('*' + self.ext), 1):
                if not os.access(child, os.R_OK):
                    raise argparse.ArgumentTypeError(f"path(s) not read-accessible '{child}'")

            if count == 0:
                raise argparse.ArgumentTypeError("directory has no contents " +
                                                 (f"({self.ext}) " if self.ext else "") +
                                                 f"'{value}'")

        return path



"""Pipeline Step to build models via AutoML: Learn"""
import os
import pathlib
import typing

import argparse_formatter

from nprintml import pipeline
from nprintml.util import format_handlers, FileAccessType, NumericRangeType

from . import AutoML


class LearnResult(typing.NamedTuple):
    """Pipeline Step results for Learn"""
    graphs_path: pathlib.Path
    models_path: pathlib.Path


class Learn(pipeline.Step):
    """Extend given `ArgumentParser` with the AutoML interface and
    build models from labeled `features`.

    Returns a `LearnResult`.

    """
    __provides__ = LearnResult
    __requires__ = ('features',)

    def __init__(self, parser):
        group_parser = parser.add_argument_group(
            "automl",
        )

        group_parser.add_argument(
            '--test-size', '--test_size',
            default=AutoML.TEST_SIZE,
            metavar='FLOAT',
            type=NumericRangeType(float, (0, 1)),
            help="proportion of data split out for testing "
                 "(float between zero and one defaulting to %(default)s)",
        )
        group_parser.add_argument(
            '--metric',
            dest='eval_metric',
            default=AutoML.EVAL_METRIC,
            choices=AutoML.EVAL_METRICS_ALL,
            metavar='{...}',
            help="metric by which predictions will be evaluated (default: %(default)s)\n"
                 "    select from: %(choices)s",
        )
        group_parser.add_argument(
            '-q', '--quality',
            default=AutoML.QUALITY,
            choices=range(len(AutoML.QUALITY_PRESETS)),
            metavar='INTEGER',
            type=int,
            help="model fit quality level (default: %(default)s)\n"
                 "    select from: " +
                 ', '.join(f'{label} ({index})'
                           for (index, label) in enumerate(AutoML.QUALITY_PRESETS)),
        )
        group_parser.add_argument(
            '--limit',
            default=AutoML.TIME_LIMIT,
            dest='time_limit',
            metavar='INTEGER',
            type=int,
            help="maximum time (seconds) over which to train each model "
                 "(default: %(default)s)",
        )

        # below %(prog)s will refer to subcommand -- not what we want
        base_program = 'python -m nprintml' if parser.prog.endswith('.py') else parser.prog

        learn_only_parser = parser.subparsers.add_parser(
            'learn',
            description="run only the AutoML step given previously-saved or "
                        "separately-prepared feature data\n\n"
                        "arguments related to AutoML are optionally inherited from the base "
                        "command (e.g.: --metric):\n\n"
                        f"    {base_program} --metric=f1 learn ./data/features.fhr",
            formatter_class=argparse_formatter.FlexiFormatter,
            help="run only AutoML",
            satisfies=self.__requires__,
        )
        learn_only_parser.add_argument(
            'features_file',
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="path to features",
        )
        learn_only_parser.set_defaults(
            __subparser__=learn_only_parser,
        )

        parser.set_defaults(
            features_file=None,
        )

    def __pre__(self, parser, args, results):
        if args.features_file:
            try:
                reader = format_handlers.get_reader(args.features_file)
            except NotImplementedError:
                subparser = args.__subparser__
                subparser.error(f"unsupported file type: {args.features_file}")

            results.features = reader(args.features_file)

    def __call__(self, args, results):
        learn = AutoML(results.features, args.outdir / 'model')

        learn(
            test_size=args.test_size,
            eval_metric=args.eval_metric,
            quality=args.quality,
            time_limit=args.time_limit,
            verbosity=args.verbosity,
        )

        return LearnResult(
            graphs_path=learn.graphs_path,
            models_path=learn.models_path,
        )



"""Pipeline Step to aggregate nPrint data under labels appropriate for
machine learning: Label

"""
import argparse
import textwrap
import typing

import pandas as pd

from nprintml import pipeline
from nprintml.util import format_handlers, HelpAction

from .aggregator import registry as aggregators


class LabelResult(typing.NamedTuple):
    """Pipeline Step results for Label"""
    features: pd.DataFrame


class Label(pipeline.Step):
    """Extend given `ArgumentParser` with the label-aggregation
    interface and invoke the requested `LabelAggregator` to extract
    features from nPrint data for the subsequent pipeline step of
    machine learning.

    Returns a `LabelResult`.

    """
    __pre_provides__ = ('labels',)
    __provides__ = LabelResult
    __requires__ = ('nprint_stream',)

    feature_file_formats = (
        'csv', 'csv.gz',
        'parquet.brotli', 'parquet.gzip', 'parquet.snappy',
        'feather.lz4', 'feather.zstd',
    )
    feature_file_formats_default = 'feather.zstd'

    def __init__(self, parser):
        self.group_parser = parser.add_argument_group(
            "aggregation of data under supplied labels",
        )

        self.group_parser.add_argument(
            '-L', '--label-file', '--label_file',
            metavar='FILE',
            required=True,
            type=argparse.FileType('r'),
            help="label file (CSV)",
        )
        self.group_parser.add_argument(
            '-a', '--aggregator',
            choices=aggregators,
            required=True,
            help="label aggregation method",
        )
        self.group_parser.add_argument(
            '--help-aggregator',
            action=HelpAction,
            help_action=print_aggregators,
            help="describe aggregators and exit",
        )
        self.group_parser.add_argument(
            '--sample-size', '--sample_size',
            metavar='INTEGER',
            type=int,
            default=1,
        )
        self.group_parser.add_argument(
            '--compress',
            action='store_true',
            help="drop columns which do not appear to provide any predictive signal",
        )
        self.group_parser.add_argument(
            '--no-save-features',
            action='store_false',
            dest='save_features',
            help="disable writing of features to disk "
                 "(writing enabled by default for inspection & reuse)",
        )
        self.group_parser.add_argument(
            '--save-features-format',
            choices=self.feature_file_formats,
            default=self.feature_file_formats_default,
            help="file format in which to save features on disk (default: %(default)s)",
        )

        self.aggregator = None

    def __pre__(self, parser, args, results):
        aggregator_class = aggregators[args.aggregator]
        self.aggregator = aggregator_class(args.label_file)

        results.labels = self.aggregator.labels

    def __call__(self, args, results):
        features = self.aggregator(
            results.nprint_stream,
            compress=args.compress,
            sample_size=args.sample_size,
        )

        if args.save_features:
            writer = format_handlers.get_writer(args.save_features_format)

            outdir = args.outdir / 'feature'
            outdir.mkdir()

            writer(features, outdir)

        return LabelResult(features)


def print_aggregators(parser, _namespace, _values, _option_string):
    """Document the aggregators by printing their doc strings."""
    print(f"{parser.prog}: labeling data aggregators")

    for (aggregator_name, aggregator_class) in aggregators.items():
        print('\n', aggregator_name, ':', sep='')

        # documentation
        doc = aggregator_class.__doc__ or aggregator_class.__init__.__doc__

        # fix indentation
        #
        # first line may not be indented at all
        if doc[0] in (' ', '\n'):
            doc = textwrap.dedent(doc)
        else:
            line_end = doc.find('\n') + 1
            if line_end > 1:
                doc = doc[:line_end] + textwrap.dedent(doc[line_end:])

        print(textwrap.indent(doc.strip(), '  '))



import itertools
import pathlib

import numpy as np
import pandas as pd

from nprintml.util import prime_iterator, storeresults

from . import LabelAggregator, AggregationLengthError, AggregationPathError


class PcapLabelAggregator(LabelAggregator):
    """A LabelAggregator that takes a directory of nPrint output files
    and considers each file a single sample, padding to the maximum
    size, to run autoML on the generated samples.

    """
    @classmethod
    def normalize_npt(cls, npt_csv, path_input_base=None):
        if isinstance(npt_csv, str):
            yield from cls.normalize_npt(pathlib.Path(npt_csv))
            return

        if isinstance(npt_csv, pathlib.Path):
            base_path = path_input_base or npt_csv
            file_count = 0
            for (file_count, npt_file) in enumerate(npt_csv.rglob('*'), 1):
                if npt_file.is_file():
                    index = str(npt_file.relative_to(base_path).with_suffix(''))
                    yield (index, npt_file)

            if file_count == 0:
                raise AggregationPathError(
                    f"{cls.__name__} requires at least one nPrint "
                    f"data file but directory is empty: '{npt_csv}'"
                )

            return

        # treat as file object stream
        file_count = 0
        for (file_count, npt_file) in enumerate(npt_csv, 1):
            file_path = pathlib.Path(npt_file.name)
            if path_input_base:
                file_path = file_path.relative_to(path_input_base)
            index = str(file_path.with_suffix(''))
            yield (index, npt_file)

        if file_count == 0:
            raise AggregationLengthError(
                f"{cls.__name__} requires at least one nPrint "
                f"data result but stream was empty: {npt_csv}"
            )

    @staticmethod
    @storeresults
    def flatten_npt(indexed_files):
        print('Loading nPrints')

        # Header should be constant across all files so configure based on first.
        #
        # Must establish what the header is (cols_header: to share) and
        # how long it is / its indices (use_cols: internal).
        try:
            indexed_files = prime_iterator(indexed_files)
        except StopIteration:
            # whoops already empty
            cols_header = use_cols = None
        else:
            (_file_index, npt_file) = indexed_files.first

            header = np.genfromtxt(npt_file, delimiter=',', max_rows=1, dtype=str)

            # ignore data index (src_ip)
            cols_header = header[1:]
            use_cols = np.arange(1, len(header))

            # be kind -- rewind (the first file descriptor)
            try:
                file_seek = npt_file.seek
            except AttributeError:
                # might be str, Path or something else we don't understand
                pass
            else:
                file_seek(0)

        # We need every sample size to be the same, so we'll report the
        # length of the largest nPrint in all of the files, such that
        # each sample may be padded with nPrint-appropriate empty values.
        max_length = file_count = 0

        for (file_count, (file_index, npt_file)) in enumerate(indexed_files, 1):
            npt = np.genfromtxt(
                npt_file,
                delimiter=',',
                skip_header=1,
                usecols=use_cols,
                dtype=int,
            )

            npt_dim1 = len(npt.shape) == 1
            npt_length = 1 if npt_dim1 else len(npt)
            max_length = max(max_length, npt_length)

            npt_flat = npt if npt_dim1 else npt.ravel()

            yield list(itertools.chain([file_index], npt_flat))

        print('Loaded', file_count, 'nprints')

        return (cols_header, max_length)

    @classmethod
    def merge_npt(cls, npts_flat):
        """Merge nPrint data from multiple output files."""
        # note: specification of dtype here is ineffective
        npts = pd.DataFrame(npts_flat)

        # true index is generated from input; rather than iterate
        # multiple times, it's initially treated as data.
        # now we can move it in place:
        npts.set_index(0, inplace=True)

        # both columns and indices were ranges and as such the name of the
        # index is currently set to the integer 0.
        # for clarity and compatibility (e.g. with output to Feather) let's
        # give the index its proper name:
        npts.index.name = 'pcap'

        # npt lengths are unknown ahead of time; and, with flatten(), each becomes a row length.
        # luckily, we need not make them equal first -- pandas fills in na or None.
        # constructor doesn't allow specification of this missing sentinel;
        # here we set it:
        npts.fillna(-1, inplace=True)

        # dtype is *mostly* int8 (byte) --
        # *except* might include relative timestamp
        npts = npts.apply(pd.to_numeric, downcast='integer')

        # though column names might not be TOO important they may be helpful.
        # here we flatten these as well from the maximum size we might require:
        (header, max_length) = npts_flat.result
        if header is not None:
            npts.columns = cls.flatten_columns(header, max_length)

        print('nPrints padded to maximum size:', max_length)
        print('nPrint features',
              'shape:', npts.shape,
              'dtypes:', npts.dtypes.unique(),
              'size (mb):', npts.memory_usage(deep=True).sum() / 1024 / 1024)

        return npts

    def __init__(self, label_csv):
        super().__init__(label_csv)
        self.labels = self.load_label(label_csv)

    def __call__(self, npt_csv, compress=False, sample_size=1, path_input_base=None):
        """Enumerate given stream or directory of nPrint data, load,
        pad to their maximum size, and attach labels.

        `npt_csv` may specify either a path to a directory of nPrint
        results (`str` or `pathlib.Path`) OR a stream of open file-like
        objects (exposing both `read()` and `name`).

        `path_input_base` is suggested when `npt_csv` specifies a stream
        of file objects, to indicate their common base path (even if
        this is virtual), such that they may be matched with the label
        index.

        """
        indexed_files = self.normalize_npt(npt_csv, path_input_base)

        npts_flat = self.flatten_npt(indexed_files)

        npt_frame = self.merge_npt(npts_flat)

        if compress:
            print('Compressing nPrint')
            npt_frame = self.compress_npt(npt_frame)
            print('  compressed nPrint shape:', npt_frame.shape)

        print('Attaching labels to nPrints')
        labels = self.prejoin_label()

        (features, missing_labels, ogns, nns) = self.attach_label(npt_frame, labels)

        print('  labels attached: missing labels for:', missing_labels)
        print('    missing labels caused samples to be dropped:', (ogns - nns))

        return features

    def prejoin_label(self):
        # Enable join on file path index with .npt-derived data --
        # strip any file suffix like .pcap
        index = self.labels.index.str.replace(r'\.pcap$', '', case=False)
        return self.labels.set_axis(index)



"""nprintML command-line interface entry-point"""
import argparse
import itertools
import multiprocessing
import os
import pathlib
import re
import sys
import time
import toml

import argparse_formatter

import nprintml
from nprintml.pipeline import Pipeline
from nprintml.util import NumericRangeType

# ensure default steps of pipeline auto-load
import nprintml.net.step
import nprintml.label.step
import nprintml.learn.step


ANIMALS = ('aardvark', 'bison', 'canary', 'dalmation', 'emu', 'falcon', 'gnu',
           'hamster', 'impala', 'jellyfish', 'kiwi', 'lemur', 'manatee',
           'nutria', 'okapi', 'porcupine', 'quetzal', 'roadrunner', 'seal',
           'turtle', 'unicorn', 'vole', 'wombat', 'xerus', 'yak', 'zebra')


PROGRAM_DESCRIPTION = """\
train models for network traffic analysis

executes a pipeline to:

  1. extract network traffic data via nPrint
  2. aggregate these data under supplied labels
  3. generate models of these labeled data via AutoGluon (AutoML)

general (optional) arguments and arguments specific to the above steps of the pipline follow below.

to execute only one portion of the pipeline -- (e.g. to run only the AutoML step given previously-\
saved or separately-prepared feature data) -- see the "subcommands" which follow (namely "learn").
"""


def execute(argv=None, **parser_kwargs):
    """Execute the nprintml CLI command."""
    args = None

    try:
        parser = build_parser(**parser_kwargs)

        pipeline = Pipeline(parser)

        finalize_parser(parser)

        args = Namespace.build(pipeline)

        parser.parse_args(argv, args)

        check_output_directory(args)

        for (step, results) in pipeline(parser, args, meta={}):
            print(step, results, sep=' → ')

        with args.meta_file.open('w') as meta_file:
            dump_meta(pipeline.results, meta_file)

        print('done →', pipeline.results)
    except KeyboardInterrupt:
        print('interrupted ✕')
    except Exception as exc:
        if args is not None and args.traceback:
            raise

        print(f'error:{exc_repr(exc)} ✕')
        sys.exit(1)


def build_parser(**parser_kwargs):
    """Construct base parser & arguments for nprintml.

    Link back to parser is set on any resulting parsed argument
    `Namespace` as `__parser__`.

    """
    parser = ArgumentParser(
        description=PROGRAM_DESCRIPTION,
        formatter_class=argparse_formatter.FlexiFormatter,
        **parser_kwargs,
    )

    version = f'nprintML {nprintml.__version__} | nPrint {nprintml.__nprint_version__}'
    # support published command aliases
    prog = pathlib.Path(sys.argv[0])
    if prog.name != 'nprintml' and prog.suffix != '.py':
        version = f'%(prog)s {nprintml.__version__} | {version}'

    parser.add_argument(
        '--version',
        action='version',
        help="show program version and exit",
        version=version,
    )

    parser.add_argument(
        '-Q', '--quiet',
        dest='verbosity',
        action='store_const',
        const=0,
        help="minimal output verbosity",
    )
    parser.add_argument(
        '-V', '--verbose',
        dest='verbosity',
        action='store_const',
        const=2,
        help="increased output verbosity",
    )
    parser.add_argument(
        '-VV', '--very-verbose',
        dest='verbosity',
        action='store_const',
        const=3,
        help="high output verbosity (e.g. print human readable packets with nPrints)",
    )
    parser.add_argument(
        '-VVV', '--debug',
        dest='verbosity',
        action='store_const',
        const=4,
        help="highest output verbosity",
    )

    parser.add_argument(
        '--tb', '--traceback',
        action='store_true',
        dest='traceback',
        help="print exception tracebacks",
    )

    try:
        # glibc-only
        sched_getaffinity = os.sched_getaffinity
    except AttributeError:
        # Note: this *may* be inaccurate in some shared environs
        cpu_available_count = multiprocessing.cpu_count()
    else:
        cpu_available_count = len(sched_getaffinity(0))

    parser.add_argument(
        '--concurrency',
        default=cpu_available_count,
        metavar='INTEGER',
        type=NumericRangeType(int, (0, None)),
        help="maximum number of concurrent processes to apply to data preparation "
             "(defaults to number reported by scheduler: %(default)s)",
    )

    parser.add_argument(
        '-o', '--output',
        default=get_default_directory(),
        dest='outdir',
        metavar='DIR',
        type=pathlib.Path,
        help="output directory path to which to write artifacts and results "
             "(default: %(default)s)",
    )

    parser.add_subparsers(
        title="subcommands",
        help="commands customizing pipeline execution",
        action=ExclusiveStepAction,
        dest='subcommand',
        resolver='__pre_satisfy__',
    )

    parser.set_defaults(
        verbosity=1,
        __parser__=parser,
    )

    return parser


def finalize_parser(parser):
    # cosmetic: ensure "subcommands" group listed at end of --help
    #
    # (required to have been defined for extension by pipeline steps;
    # but section makes most sense to be documented at very end).
    #
    action_groups = parser._action_groups
    subcommands_group = action_groups.pop(2)
    assert subcommands_group.title == 'subcommands'
    action_groups.append(subcommands_group)


class ArgumentParser(argparse.ArgumentParser):

    @property
    def subparsers(self):
        actions = self._subparsers._group_actions

        if not actions:
            return None

        (action,) = actions
        return action


class Namespace(argparse.Namespace):

    meta_file_name = 'meta.toml'

    @classmethod
    def build(cls, pipeline):
        return cls(
            __pre_satisfy__=pipeline.pre_satisfy,
        )

    @property
    def meta_file(self):
        return self.outdir / self.meta_file_name


class ExclusiveStepAction(argparse._SubParsersAction):

    def __init__(self,
                 option_strings,
                 prog,
                 parser_class,
                 dest,      # unlike parent 'dest' required
                 resolver,  # added parameter 'resolver'
                 required=False,
                 help=None,
                 metavar=None):
        # "required" added in Py37
        if sys.version_info >= (3, 7):
            super().__init__(option_strings, prog, parser_class, dest, required, help, metavar)
        elif not required:
            super().__init__(option_strings, prog, parser_class, dest, help, metavar)
        else:
            raise TypeError("keyword 'required' not supported prior to Python v3.7")

        self._name_satisfies_map = {}
        self._resolver = resolver

    def add_parser(self, name, *, satisfies=(), **kwargs):
        self._name_satisfies_map[name] = (satisfies,) if isinstance(satisfies, str) else satisfies
        return super().add_parser(name, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        super().__call__(parser, namespace, values, option_string)

        # the user has invoked an "exclusive step" subcommand -- we'll:
        #
        # 1. inform (the pipeline) of the specially-satisfied
        #    requirements associated with this case (s.t. it may update
        #    itself)
        #
        # 2. based on those steps the pipeline reports ejected:
        #
        #    2a. mark the arguments of ejected steps not required
        #    2b. raise conflict errors for *any* of these arguments
        #        that are supplied
        #
        # in so doing we'll both dynamically customize the pipeline and
        # the CLI based on the invocation of this subcommand.

        # argparse keeps nice track of what it's seen and what it hasn't
        # from within _parse_known_args (and which is ostensibly the
        # caller of this method); however, short of reaching into there,
        # we can do s'thing simple & robust here to check for conflicts.
        #
        # when checking for conflicts, argparse itself treats an
        # argument holding its default value as having not been supplied
        # -- even though these situations merely result in equivalent
        # argumentation (i.e. the user may have supplied the argument
        # with its default). nonetheless, we'll do the same.

        subcommand = getattr(namespace, self.dest)
        satisfied = self._name_satisfies_map[subcommand]
        resolver = getattr(namespace, self._resolver)

        for removed_group in resolver(*satisfied):
            actions = removed_group._group_actions if removed_group else ()
            for action in actions:
                if (
                    action.default is not argparse.SUPPRESS and
                    getattr(namespace, action.dest) is not action.default
                ):
                    action_name = argparse._get_action_name(action)
                    raise argparse.ArgumentError(None, f'argument {subcommand}: '
                                                       f'not allowed with argument {action_name}')

                action.required = False


def exc_repr(exc):
    """Construct representation of given exception appropriate for
    printed output.

    """
    exc_repr = ''

    if exc.__class__.__module__ != 'builtins':
        exc_repr += f'{exc.__class__.__module__}.'

    exc_repr += exc.__class__.__name__

    if exc.args:
        exc_repr += ': ' + ', '.join(map(str, exc.args))

    return exc_repr


def pairwise(iterable):
    """s -> (s0, s1), (s1, s2), (s2, s3), ..., (sn, None)"""
    (a, b) = itertools.tee(iterable)
    next(b, None)
    return itertools.zip_longest(a, b)


def get_default_directory(base_name='nprintml', words=ANIMALS):
    """Construct user-friendly default output directory path."""
    path = pathlib.Path(base_name)

    if path.exists():
        word_options = '|'.join(words)
        run_pattern = re.compile(rf'run-({word_options})-', re.I)
        run_matches = (run_pattern.match(run_path.name) for run_path in path.glob('run-*'))
        words_used = sorted(
            (run_match.group(1).lower() for run_match in run_matches if run_match),
            reverse=True,
        )

        if words_used:
            last_word = words_used[0]
            for (word0, word1) in pairwise(words):
                if word0 == last_word:
                    next_word = word1 or words[0]
                    break
        else:
            next_word = words[0]
    else:
        next_word = words[0]

    return path / f'run-{next_word}-{int(time.time())}-{os.getpid()}'


def check_output_directory(args):
    """Ensure output directory exists and is empty."""
    if args.outdir.exists():
        if args.outdir.is_dir():
            if any(args.outdir.iterdir()):
                args.__parser__.error(f'output directory non-empty: {args.outdir}')
        else:
            args.__parser__.error(f'output path exists and is not a directory: {args.outdir}')
    else:
        args.outdir.mkdir(parents=True)


def dump_meta(results, meta_file):
    step_timing = (
        (step.__name__, tuple(timing))
        for (step, timing) in results.__timing_steps__.items()
    )
    meta_timing = dict(step_timing, total=tuple(results.__timing__))

    meta = dict(results.meta, timing=meta_timing)

    toml.dump(meta, meta_file)



import functools
import itertools


class PrimedIterator:

    class Sentinel:
        pass

    Empty = Sentinel()
    Missing = Sentinel()

    @classmethod
    def prime(cls, iterator, default=Missing):
        """Retrieve the first item from the given iterator and return a new
        iterator over *all* elements.

        No elements from the original iterator are missed by the returned
        iterator.

        Rather, the iterator is thereby initiated, and tested for emptiness.

        This can be a useful helper in emptiness checks (which must then
        construct this secondary iterator wrapper over the first, retrieved
        element and the remainder).

        Much the same, this is useful to operations interested in the
        first element of an iterator, (but which must also iterate over
        its entirety).

        This may also be useful in initializing complex generator functions
        (e.g. those with buffers), without retrieving more than one item
        from its stream, and maintaining its full contents.

        If a second argument, the default, is given, and the iterator is
        already empty/exhausted, then this default is returned instead of
        raising StopIteration (as with `next`).

        """
        try:
            first = next(iterator)
        except StopIteration:
            if default is not cls.Missing:
                return default

            raise

        return cls(first, iterator)

    def __init__(self, first, iterator):
        self.first = first
        self.iterator = iterator
        self.__chain__ = itertools.chain((first,), iterator)

    def __iter__(self):
        yield from self.__chain__

    def __repr__(self):
        return f'({self.__class__.__name__}: {self.iterator!r})'


prime_iterator = PrimedIterator.prime


class ResultsIterator:

    @classmethod
    def storeresults(cls, generator):
        @functools.wraps(generator)
        def wrapped(*args, **kwargs):
            iterator = generator(*args, **kwargs)
            return cls(iterator)

        return wrapped

    def __init__(self, iterator):
        self.iterator = iterator
        self.result = None

    def __iter__(self):
        self.result = yield from self.iterator


storeresults = ResultsIterator.storeresults



"""nprintml"""
import pathlib

from pkg_resources import parse_requirements
from setuptools import find_packages, setup


DIR_PATH = pathlib.Path(__file__).parent

README_PATH = DIR_PATH / 'README.md'

REQUIREMENTS_PATH = DIR_PATH / 'requirement'

INSTALL_REQUIRES = [
    'argparse-formatter ~= 1.4',
    'Dickens ~= 1.0.1',

    'autogluon.tabular ~= 0.1.0',

    # autogluon.core *also* requires numpy, so match autogluon
    'numpy == 1.19.5',

    'pandas ~= 1.1.5',
    'pyarrow ~= 4.0.0',

    'matplotlib ~= 3.3.4',
    'seaborn ~= 0.11.1',

    'toml ~= 0.10.2',
]

with (REQUIREMENTS_PATH / 'dev.txt').open() as dev_fd:
    _DEV_REQUIRES = list(map(str, parse_requirements(dev_fd)))

_TESTS_REQUIRE = [
    'tox==3.23.0',
]

EXTRAS_REQUIRE = {
    'dev': _DEV_REQUIRES + _TESTS_REQUIRE,
    'test': _TESTS_REQUIRE,
}


setup(name='nprintml',
      version='1.1.2',
      description='Fully automated traffic analysis with nPrint',
      long_description=README_PATH.read_text(),
      long_description_content_type="text/markdown",
      url='https://github.com/nprint/nprintml',
      license='License :: OSI Approved :: Apache Software License',
      python_requires='>=3.6,<3.9',
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Telecommunications Industry',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: System :: Networking :: Monitoring',
          'Topic :: Security',
          'Topic :: Terminals',
      ],
      packages=find_packages('src'),
      package_dir={'': 'src'},
      entry_points={
          'console_scripts': [
              'nml=nprintml.cli:execute',
              'nprintml=nprintml.cli:execute',
              'nprint-install=nprintml.net.install:execute',
            ],
      })



"""nprintML pipeline framework

The nprintML pipeline consists of concrete subclasses of `Step`,
serially executed by a `Pipeline`.

Steps may customize the CLI's `ArgumentParser` to ensure they receive
their required execution inputs. This is performed during construction
within the step's `__init__(parser)`.

Steps must define their executable actions, via definition of their
method `__call__(args, results)__`.

Steps may control their execution order, by specifying attribute names
of the `results` object, the values of which they require preceding
steps to determine. Required values are specified as a sequence of
attribute names at the step's class level attribute
`__requires__ = ('name0', 'name1', ...)`.

Steps may indicate the `results` attributes they will provide via class
level attribute `__provides__`, the value of which may be either a
sequence of attribute names OR a supported result class -- either a
class of type `namedtuple` or `dataclass`.

Steps may report their execution results by returning their own results
object from their `__call__`. (Generic objects and named tuples are
supported. Also consider the `dataclasses` decorator package.)

The pipeline is responsible for instantiating its steps (during its own
instantiation). Upon invocation, the pipeline returns a generator, which
invokes steps as their requirements are satisfied and merges their
results into the shared results object. Each iteration of this generator
yields a tuple consisting of the last executed step and its results (if
any).

The pipeline is itself a `set` of step instances as yet unexecuted.
Steps are removed from this set as they are invoked. The composite
results object is available as the pipeline instance attribute `results`
upon pipeline invocation.

A simple pipeline example follows:

    import argparse
    import pathlib
    import tempfile
    import typing
    import urllib.request
    from xml.etree import ElementTree as ET


    class GetDataResult(typing.NamedTuple):

        data_path: pathlib.Path


    class GetData(Step):

        __provides__ = GetDataResult

        def __init__(self, parser):
            parser.add_argument(
                '-u', '--url',
                default='https://www.lycos.com/',
                help="where's the data?",
            )

        def __call__(self, args, results):
            tempdir = tempfile.TemporaryDirectory(prefix=f'{__name__}.')
            data_path = pathlib.Path(tempdir) / 'data.html'
            urllib.request.urlretrieve(args.url, data_path)
            return GetDataResult(data_path)


    class UseDataResult(typing.NamedTuple):

        img_tags: list[ET.Element]


    class UseData(Step):

        __provides__ = UseDataResult
        __requires__ = ('data_path',)

        def __init__(self, parser):
            parser.add_argument(
                '-t', '--tag',
                default='img',
                help="tag type to extract",
            )

        def __call__(self, args, results):
            root = ET.parse(results.data_path)
            imgs = root.findall('.//img')
            return UseDataResult(imgs)


    if __name__ == '__main__':
        parser = argparse.ArgumentParser(description="scrape a web site")

        # pass base parser to Pipeline construction
        # (and thereby to Step constructors):
        pipeline = Pipeline(parser)

        args = parser.parse_args()

        # can exhaust pipeline iterator however --
        # e.g., gather step results into list or handle loop:
        for (step, result) in pipeline(args):
            print(step, result)

        # results collected on pipeline.results:
        print(pipeline.results)

In the above example, steps were automatically placed into a process-
global registry, from which the pipeline built itself by default.

In fact, any `set` may serve as a step registry, and a step may be added
to any number of registries. A step may be manually added to a registry
via `register`:

    registry = set()

    GetData.register(registry)

Alternatively, a step may be added to a custom registry upon
declaration:

    class SayHello(Step, registry=registry):

        def __call__(self, args, results):
            print('Hello')

And the pipeline may be contructed from any registry of steps:

    pipeline = Pipeline(parser, registry)

(The Step and Pipeline default registries may be overridden at the class/
metaclass level, via `__default_registry__`.)

"""
import abc
import collections
import itertools
import time
import types

from descriptors import classproperty


#: Package default registry of concrete Steps
__step_registry__ = set()


class StepMeta(abc.ABCMeta):
    """Metaclass for `Step`."""
    #
    # Note that this metaclass doesn't do much of anything that *couldn't* be
    # handled on Step, (thanks to the modern hook __init_subclass__).
    #
    # However, abc entails a custom metaclass, either implicitly or explicitly;
    # and, this allows for cleaner encapsulation of class-specific
    # functionality, and a cleaner interface (as the metaclass interface is made
    # available to the class but *not* to the instance).
    #

    #: Registry to which concrete Steps will be added by default
    __default_registry__ = __step_registry__

    def register(cls, registry):
        """Add the Step to the given registry."""
        registry.add(cls)


class Step(metaclass=StepMeta):
    """Step of a Pipeline, specifying its command-line interface,
    executive functionality, and requirements from preceding Step(s).

    Steps are added to their `__default_registry__` by default upon
    declaration. A Step may be added to an alternate registry via
    declaration argument `registry`, e.g.:

        registry = set()

        class StepOne(Step, registry=registry):

            ...

    """
    #: Sequence of data keys to be set on the results object or returned
    #: for merging into the results object OR a supported Step results
    #: class from which these may be determined.
    #: Allows Pipeline to ensure that other Steps' __requires__ are met.
    #: Sequences of keys, namedtuples and dataclasses are supported.
    __provides__ = ()

    __pre_provides__ = ()

    #: Sequence of data keys on the results object -- gathered from the
    #: execution of preceding steps -- required for this step to execute
    __requires__ = ()

    @classproperty
    def provides(cls):
        try:
            return cls.__provides__._fields
        except AttributeError:
            pass

        try:
            return cls.__provides__.__dataclass_fields__
        except AttributeError:
            pass

        return cls.__provides__

    @classmethod
    def __init_subclass__(cls, *, registry=None, **kwargs):
        """Register concrete Step upon declaration."""
        super().__init_subclass__(**kwargs)

        if getattr(cls, '__abstractmethods__', None):
            if registry is not None:
                raise TypeError(
                    f"Can't register abstract class {cls.__name__}. "
                    "To customize inherited default, override class attribute "
                    f"__default_registry__ (preferably on the metaclass {StepMeta.__name__}"
                )
        else:
            cls.register(cls.__default_registry__ if registry is None else registry)

    def __init__(self, parser):
        """Initialize Step & extend given `ArgumentParser` with Step's
        command-line interface.

        """

    def __pre__(self, parser, args, results):
        """Execute preparatory actions.

        Invoked before pipeline is run and Steps' main functionality
        (`__call__`) is invoked.

        """

    @abc.abstractmethod
    def __call__(self, args, results):
        """Step's executive functionality.

        Args:
            args: parsed argument `Namespace`
            results: pipeline execution's composite results object
                (including results collected from all preceding steps)

        Return a structured data object to report the results of Step
        execution and to share these with subsequent Steps.

        """

    def __repr__(self):
        return 'step:' + self.__class__.__name__

    @property
    def __name__(self):
        return self.__class__.__name__.lower()


class Pipeline(list):
    """Pipeline of executable Steps, instantiated from and extending an
    `ArgumentParser`, and executed upon iterative invocation.

    A `Pipeline` may instantiate any iterable of Steps. By default,
    those Steps registered under the pipeline's `__default_registry__`
    are used.

    The pipeline is itself a `list` of instantiated steps to be
    executed. Step execution order is determined upon instantiation,
    according to `Step.__provides__` and `Step.__requires__`. If step
    requirements cannot be met by a given pipeline,
    `StepRequirementError` is raised.

    Steps are removed from the pipeline collection as they are executed.

    Result objects returned by steps are merged with a composite
    pipeline result object, accessible at instance attribute `results`.

    A pipeline is executed by iterating over the result of its
    invocation -- either looping over this iterator or collecting it
    into a sequence.

    For example:

        parser = argparse.ArgumentParser(description="pipeline demo")
        pipeline = Pipeline(parser)
        args = parser.parse_args()

        # can exhaust pipeline iterator however --
        # e.g., gather step results into list or handle loop:
        for (step, result) in pipeline(args):
            print(step, result)

        # results collected on pipeline.results:
        print(pipeline.results)

    If neither iterative inspection nor collection of step results is
    desired, instance method `Pipeline.exhaust(args, results=None)`
    may be invoked rather than invoking the pipeline itself.

    For example, altering the above:

        pipeline.exhaust(args)

        print(pipeline.results)

    """
    #: Registry of Steps from which Pipeline will be constructed by default
    __default_registry__ = __step_registry__

    #: Default constructor for composite results object
    results_class = types.SimpleNamespace

    @staticmethod
    def resolve(registry):
        """Generate Steps from the given `registry` in their requisite
        order.

        Step dependencies are determined from Steps' `__provides__` and
        `__requires__`.

        """
        # input registry may be any iterable
        steps = set(registry)
        results = set(itertools.chain.from_iterable(step.__pre_provides__ for step in steps))
        run = []

        while steps:
            steps_satisfied = [step for step in steps if results.issuperset(step.__requires__)]

            if not steps_satisfied:
                raise StepRequirementError(run, steps, results)

            for step in steps_satisfied:
                results.update(step.provides)
                run.append(step)
                steps.remove(step)
                yield step

    def __init__(self, parser, registry=None):
        """Construct a Pipeline of Steps extending the given
        `ArgumentParser`.

        """
        if registry is None:
            registry = self.__default_registry__

        super().__init__(step(parser) for step in self.resolve(registry))

        self.results = None

    def __call__(self, parser, args, results=None, **kwargs):
        """Construct an iterator to execute the steps of the pipeline."""
        self.results = self.results_class() if results is None else results
        self.results.__dict__.update(kwargs)

        self.results.__timing__ = Timing()
        self.results.__timing_steps__ = {}

        with self.results.__timing__:
            # pre-pipline
            for step in self:
                step.__pre__(parser, args, self.results)

            run = []

            for step in self[:]:
                results_available = set(self.results.__dict__.keys())

                if not results_available.issuperset(step.__requires__):
                    raise StepRequirementError(run, self, results_available)

                step_timing = self.results.__timing_steps__[step] = Timing()

                with step_timing:
                    step_results = step(args, self.results)

                if step_results is not None:
                    self.merge(step_results)

                run.append(step)
                self.remove(step)

                yield (step, step_results)

    def exhaust(self, args, results=None):
        """Execute the steps of the pipeline (non-iteratively)."""
        iterator = self(args, results)
        collections.deque(iterator, maxlen=0)

    def merge(self, results):
        """Copy the attributes of the given results object into the
        pipeline's composite results.

        """
        try:
            data = results.__dict__
        except AttributeError:
            data = results._asdict()

        for (key, value) in data.items():
            if hasattr(self.results, key):
                raise KeyError('result key already defined', key)

            setattr(self.results, key, value)

    def pre_satisfy(self, *fields):
        """Remove from the pipeline those steps which provide the given
        pre-satisfied `fields` of results.

        This is a generator method, iteration of which produces the
        values of removed steps' `group_parser` attribute, for the
        purpose of updating the CLI to match (e.g. to ensure that
        arguments of ejected steps are not required).

        The pipeline is considered in reverse. Steps which provide no
        more than those fields which have been pre-satisfied are
        removed. Subsequently, fields which were required by only these
        ejected steps are considered pre-satisfied as well (by cascade).

        """
        if not fields:
            return

        satisfied = set(fields)

        for step in reversed(self):
            if satisfied.issuperset(step.provides):
                # we no longer need this step to provide what's already
                # been satisfied -- remove it
                self.remove(step)

                # without the step in the pipeline its parser interface
                # is no longer important/required -- share this with the
                # caller (rather than attempt to handle here).
                yield getattr(step, 'group_parser', None)

                # let's also pretend therefore that all of that step's
                # requirements have been satisfied -- IFF no other steps
                # have the same requirement(s)
                step_required = set(step.__requires__)
                remaining_required = itertools.chain.from_iterable(step.__requires__
                                                                   for step in self)
                cascade_satisfied = step_required.difference(remaining_required)

                satisfied.update(cascade_satisfied)


class PipelineError(Exception):
    pass


class StepRequirementError(PipelineError):

    message = "step requirements could not be satisfied"

    def __init__(self, steps_run, steps_remaining, results_available):
        super().__init__(
            self.message,
            steps_run,
            steps_remaining,
            results_available,
        )
        self.steps_run = steps_run
        self.steps_remaining = steps_remaining
        self.results_available = results_available


class Timing:

    class StaleTimer(ValueError):
        pass

    def __init__(self):
        self.t0 = self.t1 = self.proc0 = self.proc1 = None

    def start(self):
        if self.t0 is not None:
            raise self.StaleTimer

        self.t0 = time.time()
        self.proc0 = time.process_time()

    def stop(self):
        if self.t1 is not None:
            raise self.StaleTimer

        self.t1 = time.time()
        self.proc1 = time.process_time()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.stop()

    @property
    def time_elapsed(self):
        if self.t0 is None or self.t1 is None:
            return None

        return self.t1 - self.t0

    @property
    def proc_elapsed(self):
        if self.proc0 is None or self.proc1 is None:
            return None

        return self.proc1 - self.proc0

    def __iter__(self):
        yield self.time_elapsed
        yield self.proc_elapsed

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'

    def __str__(self):
        return (f'total time elapsed {self.time_elapsed} | '
                f'process time elapsed {self.proc_elapsed}')
